(in-package #:cl-canvas)

;;; A canvas for Lisp graphics output
;;; Wind32 functions based on http://cl-cookbook.sourceforge.net/win32.html


(hic:load-foreign-library "user32.dll" :module "user32.dll")
(hic:load-foreign-library "kernel32.dll" :module "kernel32.dll")
(hic:load-foreign-library "gdi32.dll" :module "gdi32.dll")

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; Header-file kinds of code
;

(defparameter *class-name* "CL-Canvas Window Class")


;--- Win32 SDK Constants

; Not all of these are verified by testing in code
; More are included than are used to allow play
(defconstant COLOR_WINDOW                 5)
(defconstant CS_HREDRAW                   1)
(defconstant CS_VREDRAW                   2)
(defconstant CW_USEDEFAULT       #x80000000)
(defconstant IDC_ARROW                32512)
(defconstant SW_SHOW                      5)
(defconstant WM_CLOSE            #x00000010)
(defconstant WM_SIZE             #x00000005)
(defconstant WM_DESTROY          #x00000002)
(defconstant WM_PAINT            #x0000000f)
(defconstant WM_ERASEBKGND       #x00000014)
(defconstant WM_CREATE           #x00000001)
(defconstant WM_MOUSEMOVE        #x00000200)
(defconstant WM_LBUTTONDOWN      #x00000201)
(defconstant WM_LBUTTONUP        #x00000202)
(defconstant WM_LBUTTONDBLCLK    #x00000203)
(defconstant WM_RBUTTONDOWN      #x00000204)
(defconstant WM_RBUTTONUP        #x00000205)
(defconstant WM_RBUTTONDBLCLK    #x00000206)
(defconstant WM_MBUTTONDOWN      #x00000207)
(defconstant WM_MBUTTONUP        #x00000208)
(defconstant WM_MBUTTONDBLCLK    #x00000209)
(defconstant WM_MOUSEWHEEL       #x0000020A)
(defconstant WM_KEYDOWN          #x00000100)
(defconstant WM_KEYUP            #x00000101)
(defconstant WM_SYSKEYDOWN       #x00000104)
(defconstant WM_SYSKEYUP         #x00000105)
(defconstant WS_BORDER           #x00080000)
(defconstant WS_CAPTION          #x00C00000)
(defconstant WS_CHILD            #x40000000)
(defconstant WS_DISABLED         #x08000000)
(defconstant WS_EX_CONTROLPARENT #x00010000)
(defconstant WS_EX_APPWINDOW     #x00040000)
(defconstant WS_EX_TOPMOST       #x00000008)
(defconstant WS_MAXIMIZEBOX      #x00010000)
(defconstant WS_MINIMIZEBOX      #x00020000)
(defconstant WS_OVERLAPPED       #x00000000)
(defconstant WS_POPUP            #x80000000)
(defconstant WS_SYSMENU          #x00080000)
(defconstant WS_THICKFRAME       #x00040000)
(defconstant WS_SIZEBOX WS_THICKFRAME)
(defconstant WS_VISIBLE          #x10000000)
(defconstant SRCCOPY             #x00CC0020)
(defconstant BI_RGB                       0)
(defconstant DIB_RGB_COLORS               0)

; key codes
(defconstant VK_BACK           #x08)
(defconstant VK_TAB            #x09)
(defconstant VK_CLEAR          #x0C)
(defconstant VK_RETURN         #x0D)
(defconstant VK_SHIFT          #x10)
(defconstant VK_CONTROL        #x11)
(defconstant VK_MENU           #x12)
(defconstant VK_PAUSE          #x13)
(defconstant VK_CAPITAL        #x14)
(defconstant VK_ESCAPE         #x1B)
(defconstant VK_CONVERT        #x1C)
(defconstant VK_NONCONVERT     #x1D)
(defconstant VK_ACCEPT         #x1E)
(defconstant VK_MODECHANGE     #x1F)
(defconstant VK_SPACE          #x20)
(defconstant VK_PRIOR          #x21)
(defconstant VK_NEXT           #x22)
(defconstant VK_END            #x23)
(defconstant VK_HOME           #x24)
(defconstant VK_LEFT           #x25)
(defconstant VK_UP             #x26)
(defconstant VK_RIGHT          #x27)
(defconstant VK_DOWN           #x28)
(defconstant VK_SELECT         #x29)
(defconstant VK_PRINT          #x2A)
(defconstant VK_EXECUTE        #x2B)
(defconstant VK_SNAPSHOT       #x2C)
(defconstant VK_INSERT         #x2D)
(defconstant VK_DELETE         #x2E)
(defconstant VK_HELP           #x2F)
(defconstant VK_0              #x30)
(defconstant VK_1              #x31)
(defconstant VK_2              #x32)
(defconstant VK_3              #x33)
(defconstant VK_4              #x34)
(defconstant VK_5              #x35)
(defconstant VK_6              #x36)
(defconstant VK_7              #x37)
(defconstant VK_8              #x38)
(defconstant VK_9              #x39)
(defconstant VK_A              #x41)
(defconstant VK_B              #x42)
(defconstant VK_C              #x43)
(defconstant VK_D              #x44)
(defconstant VK_E              #x45)
(defconstant VK_F              #x46)
(defconstant VK_G              #x47)
(defconstant VK_H              #x48)
(defconstant VK_I              #x49)
(defconstant VK_J              #x4A)
(defconstant VK_K              #x4B)
(defconstant VK_L              #x4C)
(defconstant VK_M              #x4D)
(defconstant VK_N              #x4E)
(defconstant VK_O              #x4F)
(defconstant VK_P              #x50)
(defconstant VK_Q              #x51)
(defconstant VK_R              #x52)
(defconstant VK_S              #x53)
(defconstant VK_T              #x54)
(defconstant VK_U              #x55)
(defconstant VK_V              #x56)
(defconstant VK_W              #x57)
(defconstant VK_X              #x58)
(defconstant VK_Y              #x59)
(defconstant VK_Z              #x5A)
(defconstant VK_SLEEP          #x5F)
(defconstant VK_NUMPAD0        #x60)
(defconstant VK_NUMPAD1        #x61)
(defconstant VK_NUMPAD2        #x62)
(defconstant VK_NUMPAD3        #x63)
(defconstant VK_NUMPAD4        #x64)
(defconstant VK_NUMPAD5        #x65)
(defconstant VK_NUMPAD6        #x66)
(defconstant VK_NUMPAD7        #x67)
(defconstant VK_NUMPAD8        #x68)
(defconstant VK_NUMPAD9        #x69)
(defconstant VK_MULTIPLY       #x6A)
(defconstant VK_ADD            #x6B)
(defconstant VK_SEPARATOR      #x6C)
(defconstant VK_SUBTRACT       #x6D)
(defconstant VK_DECIMAL        #x6E)
(defconstant VK_DIVIDE         #x6F)
(defconstant VK_F1             #x70)
(defconstant VK_F2             #x71)
(defconstant VK_F3             #x72)
(defconstant VK_F4             #x73)
(defconstant VK_F5             #x74)
(defconstant VK_F6             #x75)
(defconstant VK_F7             #x76)
(defconstant VK_F8             #x77)
(defconstant VK_F9             #x78)
(defconstant VK_F10            #x79)
(defconstant VK_F11            #x7A)
(defconstant VK_F12            #x7B)
(defconstant VK_F13            #x7C)
(defconstant VK_F14            #x7D)
(defconstant VK_F15            #x7E)
(defconstant VK_F16            #x7F)
(defconstant VK_F17            #x80)
(defconstant VK_F18            #x81)
(defconstant VK_F19            #x82)
(defconstant VK_F20            #x83)
(defconstant VK_F21            #x84)
(defconstant VK_F22            #x85)
(defconstant VK_F23            #x86)
(defconstant VK_F24            #x87)
(defconstant VK_NUMLOCK        #x90)
(defconstant VK_SCROLL         #x91)


; Aggregates 
(defconstant WS_POPUPWINDOW (logior ws_popup ws_border ws_sysmenu))
(defconstant WS_OVERLAPPEDWINDOW (logior ws_overlapped ws_caption 
                                         ws_sysmenu ws_thickframe
                                         ws_minimizebox ws_maximizebox))

(defmacro LOWORD (l)
  `(logand ,l #xffff))

(defmacro HIWORD (l)
  `(ash ,l -16))

(defmacro GET_X_LPARAM (lParam)
  `(LOWORD ,lParam))

(defmacro GET_Y_LPARAM (lParam)
  `(HIWORD ,lParam))


;--- Win32 SDK data types

(hic:def-foreign-type BOOL :unsigned-long)
(hic:def-foreign-type WORD :unsigned-short)
(hic:def-foreign-type DWORD :unsigned-long)
(hic:def-foreign-type HANDLE :unsigned-long)
(hic:def-foreign-type HBITMAP :unsigned-long)
(hic:def-foreign-type HGDIOBJ :unsigned-long)
(hic:def-foreign-type HDC :unsigned-long)
(hic:def-foreign-type HINSTANCE :unsigned-long)
(hic:def-foreign-type HMENU :unsigned-long)
(hic:def-foreign-type HWND :unsigned-long)
(hic:def-foreign-type INT :int)
(hic:def-foreign-type LONG :long)
(hic:def-foreign-type LPCTSTR *)
(hic:def-foreign-type LPSTR *)
(hic:def-foreign-type LPVOID :unsigned-long)
(hic:def-foreign-type LPFN :unsigned-long)
(hic:def-foreign-type UINT :unsigned-int)
(hic:def-foreign-type ULONG :unsigned-long)
(hic:def-foreign-type wBYTE :unsigned-char)
(hic:def-foreign-type PTR :unsigned-long)


;--- Win32 SDK structures

; POINT
(hic:def-struct POINT
                (x INT)
                (y INT))

; MSG
(hic:def-struct MSG
                (hwnd HWND)
                (wParam ULONG)
                (lParam ULONG)
                (time DWORD)
                (point POINT))

; PAINTSTRUCT
(hic:def-struct PAINTSTRUCT
                (HDC hdc)
                (fErase bool)
                (rcPaint-x uint)
                (rcPaint-y uint)
                (rcPaint-width uint)
                (rcPaint-height uint)
                (fRestore bool)
                (fIncUpdate bool)
                (rgbReserved (:array wBYTE 32)))

; WndClass
(hic:def-struct WNDCLASS
                (style UINT)
                (lpfnWndProc LPFN)
                (cbClsExtra INT)
                (cbWndExtra INT)
                (hInstance HANDLE)
                (hIcon HANDLE)
                (hCursor HANDLE)
                (hBrBackground HANDLE)
                (lpszMenuName ULONG) ;; ULONG so it can be set to null without error
                (lpszClassName LPCTSTR))

; RGBQUAD
(hic:def-struct RGBQUAD
                (rgbBlue wBYTE)
                (rgbGreen wBYTE)
                (rgbRed wBYTE)
                (rgbReserved wBYTE))

; BITMAPINFOHEADER
(hic:def-struct BITMAPINFOHEADER
                (biSize DWORD)
                (biWidth LONG)
                (biHeight LONG)
                (biPlanes WORD)
                (biBitCount WORD)
                (biCompression DWORD)
                (biSizeImage DWORD)
                (biXPelsPerMeter LONG)
                (biYPelsPerMeter LONG)
                (biClrUsed DWORD)
                (biClrImportant DWORD))

; BITMAPINFO
(hic:def-struct BITMAPINFO
                (bmiHeader BITMAPINFOHEADER)
                (bmiColors (:array RGBQUAD 1)))

; RECT
(hic:def-struct RECT
                (left :LONG)
                (top  :LONG)
                (right :LONG)
                (bottom :LONG))


;--- Win32 SDK functions

; BeginPaint
(hic:def-function "BeginPaint"
                  ((hwnd hwnd)
                   (lpPaintStruct ptr))
                  :module "user32.dll"
                  :returning HDC)

; CreateWindowExA
(hic:def-function "CreateWindowExA"
                  ((dwExStyle DWORD)
                   (lpClassName LPCTSTR)
                   (lpWindowName LPCTSTR)
                   (dwStyle DWORD)
                   (x UINT)
                   (y UINT)
                   (nWidth UINT)
                   (nHeight UINT)
                   (hwndParent HWND)
                   (hMenu HMENU)
                   (hInstance HINSTANCE)
                   (lpParam LPVOID))
                  :module "user32.dll"
                  :returning HWND)

; DefWindowProcA
(hic:def-function "DefWindowProcA"
                  ((hwnd ULONG)
                   (msg ULONG)
                   (wparam ULONG)
                   (lparam ULONG))
                  :module "user32.dll"
                  :returning ULONG)

; DispatchMessageA
(hic:def-function "DispatchMessageA"
                  ((MsgPtr PTR)) ;; We're passing a pointer but Lisp doesn't know
                  :module "user32.dll"
                  :returning ULONG)

; EndPaint
(hic:def-function "EndPaint"
                  ((hwnd HWND)
                   (lpPaintStruct PTR))
                  :module "user32.dll"
                  :returning BOOL)

; GetLastError
(hic:def-function "GetLastError"
                  ()
                  :module "user32.dll"
                  :returning DWORD)

; GetMessageA
(hic:def-function "GetMessageA"
                  ((lpMsg PTR)
                   (hwnd ULONG)
                   (MsgFiltMin ULONG)
                   (MsgFiltMax ULONG))
                  :module "user32.dll"
                  :returning BOOL)

; GetModuleHandleA
(hic:def-function "GetModuleHandleA"
                  ((lpModuleName LONG))
                  :module "kernel32.dll"
                  :returning HANDLE)

; LoadCursorA
(hic:def-function "LoadCursorA"
                  ((hInstance HANDLE)
                   (param (:unsigned :long)))
                  :module "user32.dll"
                  :returning HANDLE)

; PostQuitMessage
(hic:def-function "PostQuitMessage"
                  ((nExitCode :int))
                  :module "user32.dll"
                  :returning :void)

; RegisterClassA
(hic:def-function "RegisterClassA"
                  ((WndClass PTR))
                  :module "user32.dll"
                  :returning BOOL)

; TextOutA
(hic:def-function "TextOutA"
                  ((hdc HDC)
                   (nXStart UINT)
                   (nYStart UINT) 
                   (lpString LPCTSTR)
                   (cbString UINT))
                  :module "user32.dll"
                  :returning BOOL)

; TranslateMessage
(hic:def-function "TranslateMessage"
                  ((msg-ptr ULONG))
                  :module "user32.dll"
                  :returning BOOL)

; CreateCompatibleDC
(hic:def-function "CreateCompatibleDC"
                  ((hdc HDC))
                  :module "gdi32.dll"
                  :returning HDC)

; CreateDIBSection
(hic:def-function "CreateDIBSection"
                  ((hdc HDC)
                   (pbmi PTR)
                   (iUsage UINT)
                   (ppvBits LPVOID)
                   (hSection HANDLE)
                   (dwOffset DWORD))
                  :module "gdi32.dll"
                  :returning HBITMAP)

; SelectObject
(hic:def-function "SelectObject"
                  ((hdc HDC)
                   (hgdiobj HGDIOBJ))
                  :module "gdi32.dll"
                  :returning HGDIOBJ)

; BitBlt
(hic:def-function "BitBlt"
                  ((hdc HDC)
                   (nXDest INT)
                   (nYDest INT)
                   (nWidth INT)
                   (nHeight INT)
                   (hdcSrc HDC)
                   (nXSrc INT)
                   (nYSrc INT)
                   (dwRop DWORD))
                  :module "gdi32.dll"
                  :returning BOOL)

; DeleteDC
(hic:def-function "DeleteDC"
                  ((hdc HDC))
                  :module "gdi32.dll"
                  :returning BOOL)

; DeleteObject
(hic:def-function "DeleteObject"
                  ((hObject HGDIOBJ))
                  :module "gdi32.dll"
                  :returning BOOL)

; InvalidateRect
(hic:def-function "InvalidateRect"
                  ((hWnd HWND)
                   (lpRect (:pointer RECT))
                   (bErase BOOL))
                  :module "user32.dll"
                  :returning BOOL)

; GetClientRect
(hic:def-function "GetClientRect"
                  ((h HWND)
                   (l (* RECT)))
                  :module "user32.dll"
                  :returning LONG)

; MoveWindow
(hic:def-function "MoveWindow"
                  ((h HWND)
                   (X INT)
                   (Y INT)
                   (nWidth INT)
                   (nHeight INT)
                   (bRepaint BOOL))
                  :module "user32.dll"
                  :returning BOOL)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; Code specific to one program
;

;--- Win32 SDK Callbacks

(defun wndproc-create (hwnd)
  (declare (special *width* *height*))
  ;; adjust window size for the right client area size
  (hic:with-foreign-object (rect 'RECT)
    (GetClientRect hwnd rect)
    (let ((width (hic:get-slot-value rect 'RECT 'right))
          (height (hic:get-slot-value rect 'RECT 'bottom)))
      (let ((dx (- *width* width))
            (dy (- *height* height)))
        (MoveWindow hwnd
                    (hic:get-slot-value rect 'RECT 'left)
                    (hic:get-slot-value rect 'RECT 'top)
                    (+ *width* dx)
                    (+ *height* dy)
                    0)))))

(defun wndproc-paint (hwnd)
  (declare (special *width* *height* *framebuffer-dc*))
  (hic:with-foreign-object (ps 'PAINTSTRUCT)
    (let ((hdc (BeginPaint hwnd (hic:pointer-address ps))))
      ; Paint here
      (BitBlt hdc 0 0 *width* *height* *framebuffer-dc* 0 0 SRCCOPY)
      (EndPaint hwnd (hic:pointer-address ps)))))

(defparameter *on-mouse-move* nil)
(defparameter *on-l-button-down* nil)
(defparameter *on-l-button-up* nil)
(defparameter *on-l-button-double-click* nil)
(defparameter *on-r-button-down* nil)
(defparameter *on-r-button-up* nil)
(defparameter *on-r-button-double-click* nil)
(defparameter *on-m-button-down* nil)
(defparameter *on-m-button-up* nil)
(defparameter *on-m-button-double-click* nil)
(defparameter *on-key-down* nil)
(defparameter *on-key-up* nil)

(defmacro message-cond ((hwnd msg wparam lparam) &body conds)
  (labels ((eval-message-cond (conds hwnd msg wparam lparam)
             (if (null conds)
                 nil
               (destructuring-bind (cond-element &rest rest) conds
                 (cons (destructuring-bind (cond-command &rest cond-arguments) cond-element
                         (format t "args: ~a~%" cond-arguments)
                         (cond 
                          ((eql cond-command 'ignore-message)
                           `((= ,msg ,@cond-arguments) 0))
                          ((eql cond-command 'on-message)
                           (destructuring-bind (message function) cond-arguments
                             `((= ,msg ,message) (,function ,hwnd) 0)))
                          ((eql cond-command 'check-mouse-event)
                           (destructuring-bind (message function) cond-arguments
                             `((and ,function (= ,msg ,message))
                               (funcall ,function
                                        (GET_X_LPARAM ,lparam)
                                        (GET_Y_LPARAM ,lparam))
                               0)))
                          ((eql cond-command 'check-key-event)
                           (destructuring-bind (message function) cond-arguments
                             `((and ,function (= ,msg ,message))
                               (funcall ,function ,wparam)
                               0)))))
                       (eval-message-cond rest hwnd msg wparam lparam))))))
    `(cond ,@(eval-message-cond conds hwnd msg wparam lparam)
           (t (DefWindowProcA ,hwnd ,msg ,wparam ,lparam)))))

; WndProc -- Window procedure for the window we will create
(hic:ff-defun-callable :cdecl :long wndproc
    ((hwnd hwnd) (msg (:unsigned :long)) 
     (wparam (:unsigned :long)) (lparam (:unsigned :long)))
  (message-cond (hwnd msg wparam lparam)
    (on-message WM_PAINT wndproc-paint)
    (on-message WM_ERASEBKGND wndproc-paint)
    (on-message WM_CREATE wndproc-create)
    #+console (WM_DESTROY (PostQuitMessage 0))
    (check-mouse-event WM_MOUSEMOVE *on-mouse-move*)
    (check-mouse-event WM_LBUTTONDOWN *on-l-button-down*)
    (check-mouse-event WM_LBUTTONUP *on-l-button-up*)
    (check-mouse-event WM_LBUTTONDBLCLK *on-l-button-double-click*)
    (check-mouse-event WM_RBUTTONDOWN *on-r-button-down*)
    (check-mouse-event WM_RBUTTONUP *on-r-button-up*)
    (check-mouse-event WM_RBUTTONDBLCLK *on-r-button-double-click*)
    (check-mouse-event WM_MBUTTONDOWN *on-m-button-down*)
    (check-mouse-event WM_MBUTTONUP *on-m-button-up*)
    (check-mouse-event WM_MBUTTONDBLCLK *on-m-button-double-click*)
    (check-key-event WM_KEYDOWN *on-key-down*)
    (check-key-event WM_KEYUP *on-key-up*)
    (check-key-event WM_SYSKEYDOWN *on-key-down*)
    (check-key-event WM_SYSKEYUP *on-key-up*)))

;; unchecked framebuffer memory access
(defmacro framebuffer (index)
  `(hic:deref-array *framebuffer* :int ,index))

;; unchecked framebuffer point access
(defmacro framebuffer-point (x y)
  `(framebuffer (+ (* *width* ,y) ,x)))

(defparameter *framebuffer* nil)
(defparameter *framebuffer-bitmap* nil)
(defparameter *framebuffer-dc* nil)
(defparameter *old-bitmap* nil)

(defun create-framebuffer (width height)
  (declare (special *framebuffer* *framebuffer-bitmap* *framebuffer-dc* *old-bitmap*))
  (let ((bmi (hic:allocate-foreign-object 'BITMAPINFO)))
    (unwind-protect
        (let ((header (hic:get-slot-pointer bmi 'BITMAPINFOHEADER 'bmiHeader)))
          (setf (hic:get-slot-value header 'BITMAPINFOHEADER 'biSize) 40)
          (setf (hic:get-slot-value header 'BITMAPINFOHEADER 'biWidth) width)
          (setf (hic:get-slot-value header 'BITMAPINFOHEADER 'biHeight) (- height))
          (setf (hic:get-slot-value header 'BITMAPINFOHEADER 'biPlanes) 1)
          (setf (hic:get-slot-value header 'BITMAPINFOHEADER 'biBitCount) 32)
          (setf (hic:get-slot-value header 'BITMAPINFOHEADER 'biCompression) BI_RGB)
          (setf (hic:get-slot-value header 'BITMAPINFOHEADER 'biSizeImage) 0)
          (setf (hic:get-slot-value header 'BITMAPINFOHEADER 'biClrUsed) 0)
          (setf (hic:get-slot-value header 'BITMAPINFOHEADER 'biClrImportant) 0)
          (let* ((framebuffer-pointer (hic:allocate-foreign-object ':long))
                 (framebuffer-dc (CreateCompatibleDC 0))
                 framebuffer-bitmap)
            (setf (hic:deref-pointer framebuffer-pointer ':long) 0)
            (setf framebuffer-bitmap
                  (CreateDIBSection framebuffer-dc
                                    (hic:pointer-address bmi)
                                    DIB_RGB_COLORS
                                    (hic:pointer-address framebuffer-pointer)
                                    0
                                    0))
            (when (= framebuffer-bitmap 0)
              (error (format nil "~&CreateDIBSection failed. Error: ~a" (GetLastError))))
            (setf *framebuffer*
                  (hic:make-pointer (hic:deref-pointer framebuffer-pointer ':long) :int))
            (clear-framebuffer)
            (let ((old-bitmap (SelectObject framebuffer-dc framebuffer-bitmap)))
              (setf *framebuffer-bitmap* framebuffer-bitmap)
              (setf *framebuffer-dc* framebuffer-dc)
              (setf *old-bitmap* old-bitmap))))
      (hic:free-foreign-object bmi))))


;--- Main Functions

(defun register-class ()
  (hic:with-foreign-string (class-name *class-name*)
    (let (rslt
          (wndproc-p (hic:ff-register-callable 'wndproc)))
      (hic:with-foreign-object (wc 'WNDCLASS)
        (setf (hic:get-slot-value wc 'WNDCLASS 'style)
              (logior cs_hredraw cs_vredraw))
        (setf (hic:get-slot-value wc 'WNDCLASS 'lpfnWndProc)
              (hic:pointer-address wndproc-p))
        (setf (hic:get-slot-value wc 'WNDCLASS 'cbClsExtra) 0)
        (setf (hic:get-slot-value wc 'WNDCLASS 'cbWndExtra) 0)
        (setf (hic:get-slot-value wc 'WNDCLASS 'hInstance) (GetModuleHandleA 0))
        (setf (hic:get-slot-value wc 'WNDCLASS 'hIcon) 0)
        (setf (hic:get-slot-value wc 'WNDCLASS 'hCursor) (LoadCursorA 0 IDC_ARROW))
        (setf (hic:get-slot-value wc 'WNDCLASS 'hbrBackground) COLOR_WINDOW)
        (setf (hic:get-slot-value wc 'WNDCLASS 'lpszMenuName) 0)
        (setf (hic:get-slot-value wc 'WNDCLASS 'lpszClassName) class-name)
        (setq rslt (RegisterClassA (hic:pointer-address wc))))
      (if (/= rslt 0)
          rslt
        (error (format nil "~&RegisterClass failed in reg-class. Error: ~a" 
                       (GetLastError)))))))

(defvar *reg-class-atom* (register-class))
(defparameter *width* 0)
(defparameter *height* 0)
(defparameter *current-x* 0)
(defparameter *current-y* 0)
(defparameter *hwnd* nil)

; CreateWindow
(defun create-toplevel-window-run (window-name x0 y0 width height topmost)
  "See create-window."
  (declare (special *hwnd* *width* *height*))
  (setf *width* width)
  (setf *height* height)
  (setf *current-x* (/ *width* 2))
  (setf *current-y* (/ *height* 2))
  (create-framebuffer width height)
  (hic:with-foreign-string (class-name *class-name*)
    (hic:with-foreign-string (name window-name)
      (let ((hwnd (CreateWindowExA (if topmost WS_EX_TOPMOST 0)
                                   class-name name
                                   (logand (logior WS_VISIBLE WS_OVERLAPPEDWINDOW)
                                           (lognot WS_MAXIMIZEBOX)
                                           (lognot WS_MINIMIZEBOX)
                                           (lognot WS_SIZEBOX))
                                   x0 y0 width height
                                   0 0 (GetModuleHandleA 0) 0)))
        (when (zerop hwnd)
          (error (format nil "CreateWindow failed. GetLastError is ~a" 
                         (GetLastError))))
        (setf *hwnd* hwnd)
        #-console hwnd ;; Should return wParam in thread, what about here?
        #+console ;; Message pump the thread
        (hic:with-foreign-object (ms 'MSG)
          (do ((gm (GetMessageA (hic:pointer-address ms) 0 0 0)
                   (GetMessageA (hic:pointer-address ms) 0 0 0)))
              ((zerop gm) (hic:get-slot-value ms 'MSG 'wParam))
            (TranslateMessage (hic:pointer-address ms))
            (DispatchMessageA (hic:pointer-address ms))))))))

;; creates the canvas window and shows it
;; define mouse callbacks like this: #'(lambda (x y) (format t "x: ~a y: ~a~%" x y))
;; and key callbacks: #'(lambda (code) (format t "key-down: ~a~%" code))
(defun show-canvas (width height &key
                          caption
                          topmost
                          on-mouse-move
                          on-l-button-down
                          on-l-button-up
                          on-l-button-double-click
                          on-r-button-down
                          on-r-button-up
                          on-r-button-double-click
                          on-m-button-down
                          on-m-button-up
                          on-m-button-double-click
                          on-key-down
                          on-key-up)
  (unless caption (setf caption "CL-Canvas"))
  (setf *on-mouse-move* on-mouse-move)
  (setf *on-l-button-down* on-l-button-down)
  (setf *on-l-button-up* on-l-button-up)
  (setf *on-l-button-double-click* on-l-button-double-click)
  (setf *on-r-button-down* on-r-button-down)
  (setf *on-r-button-up* on-r-button-up)
  (setf *on-r-button-double-click* on-r-button-double-click)
  (setf *on-m-button-down* on-m-button-down)
  (setf *on-m-button-up* on-m-button-up)
  (setf *on-m-button-double-click* on-m-button-double-click)
  (setf *on-key-down* on-key-down)
  (setf *on-key-up* on-key-up)
  (let ((window-name caption))
    #-console (create-toplevel-window-run window-name 0 0 width height topmost)
    #+console (progn
		(when (null (mp:list-all-processes))
		  (mp:initialize-multiprocessing))
		(mp:process-run-function window-name nil
                                         'create-toplevel-window-run
                                         window-name
                                         0 0 width height topmost))))


(defparameter *current-color* 0)


(defmacro check-bounds (min below &rest vars)
  (let (result)
    (loop for var in vars do
          (push `(when (< ,var ,min) (setf ,var ,min)) result)
          (push `(when (>= ,var ,below) (setf ,var (1- ,below))) result))
    (push 'progn result)
    result))

;; repaints framebuffer
(defun repaint ()
  (declare (special *hwnd*))
  (InvalidateRect *hwnd* nil 0))

;; clipped pixel set
(defun set-pixel (x y color)
  (declare (special *width* *height* *framebuffer*))
  ;; use only integers
  (setf x (truncate x))
  (setf y (truncate y))
  ;; clip
  (check-bounds 0 *width* x)
  (check-bounds 0 *height* y)
  (setf (framebuffer-point x y) color))

;; clipped pixel get
(defun get-pixel (x y)
  (declare (special *width* *height* *framebuffer*))
  ;; use only integers
  (setf x (truncate x))
  (setf y (truncate y))
  ;; clip
  (check-bounds 0 *width* x)
  (check-bounds 0 *height* y)
  (framebuffer-point x y))

;; set current drawing color
(defun set-color (color)
  (setf *current-color* color))

;; clipped rect filling
(defun fill-rect (x0 y0 x1 y1)
  (declare (special *current-color* *framebuffer* *width* *height*))
  ;; use only integers
  (setf x0 (truncate x0))
  (setf y0 (truncate y0))
  (setf x1 (truncate x1))
  (setf y1 (truncate y1))
  ;; clipping
  (check-bounds 0 *width* x0 x1)
  (check-bounds 0 *height* y0 y1)
  ;; swapping
  (when (< x1 x0) (rotatef x0 x1))
  (when (< y1 y0) (rotatef y0 y1))
  ;; fill
  (loop for x from x0 below x1 do
        (loop for y from y0 below y1 do
              (setf (framebuffer-point x y) *current-color*))))

;; clipped line draw
(defun draw-line (x0 y0 x1 y1)
  (declare (special *current-color* *framebuffer* *width* *height*))
  ;; use only integers
  (setf x0 (truncate x0))
  (setf y0 (truncate y0))
  (setf x1 (truncate x1))
  (setf y1 (truncate y1))
  ;; simple clipping, should be improved with Cohen-Sutherland line clipping
  (check-bounds 0 *width* x0 x1)
  (check-bounds 0 *height* y0 y1)
  (let (x y e dx dy)
    (when (> x0 x1)
      (rotatef x0 x1)
      (rotatef y0 y1))
    (setf e 0)
    (setf x x0)
    (setf y y0)
    (setf dx (- x1 x0))
    (setf dy (- y1 y0))
    (if (>= dy 0)
        (if (>= dx dy)
            (loop for x from x0 to x1 do
                  (setf (framebuffer-point x y) *current-color*)
                  (if (< (* 2 (+ e dy)) dx)
                      (incf e dy)
                    (progn
                      (incf y)
                      (incf e (- dy dx)))))
          (loop for y from y0 to y1 do
                (setf (framebuffer-point x y) *current-color*)
                (if (< (* 2 (+ e dx)) dy)
                    (incf e dx)
                  (progn
                    (incf x)
                    (incf e (- dx dy))))))
      (if (>= dx (- dy))
          (loop for x from x0 to x1 do
                (setf (framebuffer-point x y) *current-color*)
                (if (> (* 2 (+ e dy)) (- dx))
                    (incf e dy)
                  (progn
                    (decf y)
                    (incf e (+ dy dx)))))
        (progn
          (rotatef x0 x1)
          (rotatef y0 y1)
          (setf x x0)
          (setf dx (- x1 x0))
          (setf dy (- y1 y0))
          (loop for y from y0 to y1 do
                (setf (framebuffer-point x y) *current-color*)
                (if (> (* 2 (+ e dx)) (- dy))
                    (incf e dx)
                  (progn
                    (decf x)
                    (incf e (+ dx dy))))))))))

(defun clear-framebuffer ()
  (declare (special *width* *height* *framebuffer*))
  (loop for i from 0 below (* *width* *height*) do (setf (framebuffer i) #xffffff)))

;; Logo-like interface

(defparameter *current-angle* 0)
(defparameter *pen-down* t)

(defun rotate (degree)
  (declare (special *current-angle*))
  (let ((radian (* (/ degree 180.0) pi)))
    (incf *current-angle* radian)))

(defun right (degree)
  (rotate (- degree)))

(defun left (degree)
  (rotate degree))

(defun pendown ()
  (declare (special *pen-down*))
  (setf *pen-down* t))

(defun penup ()
  (declare (special *pen-down*))
  (setf *pen-down* nil))

(defun forward (length)
  (declare (special *current-x* *current-y* *current-angle* *pen-down*))
  (let ((dx (* length (sin *current-angle*)))
        (dy (* length (cos *current-angle*))))
    (let ((x2 (+ *current-x* dx))
          (y2 (+ *current-y* dy)))
      (when *pen-down* (draw-line *current-x* *current-y* x2 y2))
      (setf *current-x* x2)
      (setf *current-y* y2)))
  (repaint))

(defun kochcurve (level size)
  (if (= level 0)
      (forward size)
    (let ((level (1- level))
          (size (/ size 3)))
      (kochcurve level size)
      (left 60)
      (kochcurve level size)
      (right 120)
      (kochcurve level size)
      (left 60)
      (kochcurve level size))))

; (snow 4 140)
(defun snow (level size)
  (kochcurve level size)
  (right 120)
  (kochcurve level size)
  (right 120)
  (kochcurve level size)
  (right 120)
  (kochcurve level size))

; (mandelbrot 0.2 0.5 0.4 0.7)
(defun mandelbrot (x0 y0 x1 y1)
  (declare (special *width* *height*))
  (loop for y from 0 below *height* do
        (loop for x from 0 below *width* do
              (loop with a = (complex
                              (float (+ (* (/ (- x1 x0) *width*) x) x0))
                              (float (+ (* (/ (- y1 y0) *height*) y) y0)))
                    for z = a then (+ (* z z) a)
                    while (< (abs z) 2)
                    for c from 60 above 0
                    finally (set-pixel x y (logior (mod (* 13 c) 256)
                                                   (ash (mod (* 7 c) 256) 8)
                                                   (ash (mod (* 2 c) 256) 16))))))
  (repaint))

(defun test ()
  (show-canvas 300 300)
  (snow 4 140))


; drawing program demo
(let ((mouse-pressed nil)
      (last-x 0)
      (last-y 0))
  (defun on-l-button-down (x y)
    (setf mouse-pressed t
          last-x x
          last-y y))
  (defun on-l-button-up (x y)
    (declare (ignore x y))
    (setf mouse-pressed nil))
  (defun on-mouse-move (x y)
    (when mouse-pressed 
      (draw-line last-x last-y x y)
      (repaint)
      (setf last-x x
            last-y y))))

(defun drawing-program ()
  (show-canvas 400 400
               :caption "Drawing program"
               :on-mouse-move #'on-mouse-move
               :on-l-button-down #'on-l-button-down
               :on-l-button-up #'on-l-button-up
               :on-key-down #'(lambda (code)
                                (when (= code VK_SPACE)
                                  (clear-framebuffer)
                                  (repaint)))))
